import {PolymerElement} from '@polymer/polymer/polymer-element.js';
import {html} from '@polymer/polymer/lib/utils/html-tag.js';
import {mixinBehaviors} from '@polymer/polymer/lib/legacy/class.js';
import {IronValidatableBehavior} from '@polymer/iron-validatable-behavior/iron-validatable-behavior.js';
import {afterNextRender} from '@polymer/polymer/lib/utils/render-status.js';
import {ApiFormMixin} from '@api-components/api-form-mixin/api-form-mixin.js';
import '@api-components/api-form-mixin/api-form-styles.js';
import '@polymer/polymer/lib/elements/dom-if.js';
import '@polymer/polymer/lib/elements/dom-repeat.js';
import '@polymer/iron-form/iron-form.js';
import '@polymer/paper-icon-button/paper-icon-button.js';
import '@advanced-rest-client/arc-icons/arc-icons.js';
import '@advanced-rest-client/arc-definitions/arc-definitions.js';
import '@polymer/paper-checkbox/paper-checkbox.js';
import './api-headers-form-item.js';
/**
 * HTTP headers form build from AMF json/ld model.
 * The form is pre-populated with predefined values from the AMF model
 * using `api-view-model-transformer` element. This element is included in
 * `api-headers-editor`.
 *
 * **This API component replaces `raml-headers-form` component**. Old
 * component uses RAML data model generated by raml JS parser only.
 * If you need RAML form only or component that works with Polymer 1.0
 * use `raml-headers-form` instead.
 *
 * ## Building the view model
 *
 * For practical reasons this element consumes different data model than the
 * one used in AMF. Use `api-view-model-transformer` element to generate
 * data view model from AMF JSON ld schema. See demo page for example of
 * use.
 *
 * ### Generating view model from AMF
 *
 * ```html
 * <api-view-model-transformer on-view-model-changed="_updateView"></api-view-model-transformer>
 * <api-headers-form model="{{viewModel}}" value="{{headers}}"></api-headers-form>
 * <script>
 * const amfModel = getAmfFromRamlOrOas();
 * const processor = document.querySelector('api-view-model-transformer');
 * if (amfModel['@context']) {
 *  processor.amfContext = amfModel['@context'];
 * }
 * processor.amfModel = extractHeadersForMethod(amfModel);
 * processor.addEventListener('view-model-changed', (e) => {
 *  document.querySelector('api-headers-form') = e.detail.value;
 * });
 * < /script>
 * ```
 *
 * This example uses `getAmfFromRamlOrOas()` function where you implement
 * the logic of getting AMF json/ld data. It can be stored in file or parsed
 * using AMF parsers. The `extractHeadersForMethod()` function implements a
 * logic of extracting headers definition from a method (for example).
 *
 * ### Manually generating view model
 *
 * You can use `api-view-model-transformer` and `buildProperty()` function
 * to generate basic view model. You can also construct it on your own.
 * See `api-view-model-transformer` element documentation for model schema.
 *
 * ## Changes in v2
 *
 * The component won't listen for `headers-value-changed` custom event.
 * Components / applications using this element should handle headers change
 * event in the application and generate new model for the view.
 * Setting `value` makes no effect on the element. Only way to change
 * generated UI is to change the model.
 *
 * ## Styling
 *
 * Custom property | Description | Default
 * ----------------|-------------|----------
 * `--api-headers-form` | Mixin applied to this element | `{}`
 * `--api-headers-form-delete-icon-margin-top` | Margin top of the delete button | `16px`
 * `--api-headers-form-delete-icon-margin-top-narrow` | Margin top of the delete button in narrow layout | `16px`
 * `--api-headers-form-enable-checkbox-margin-top` | Margin top of the enable checkbox | `32px`
 * `--api-headers-form-enable-checkbox-margin-top` | Margin top of the enable checkbox in narrow layout | `32px`
 * `--api-headers-form-item` | Mixin applied to each form item. Each for item is separate element. | `{}`
 * `--api-headers-editore-hint-icon-margin-top` | Margin top of help icon. | `16px`
 * `--api-headers-editore-hint-icon-margin-top-narrow` | Margin top of help icon in narrow layout. | `16px`
 * `--api-headers-form-name-input` | Mixin applied to header name input | `{}`
 * `--api-headers-form-name-input-narrow` | Mixin applied to header name input in narrow layout | `{}`
 * `--api-headers-form-array-parameter` | Mixin applied to the value input when the input is an array | `{}`
 *
 * See `api-headers-form-item` and `api-form-mixin/api-form-styles` for more styling API.
 *
 * @customElement
 * @polymer
 * @demo demo/index.html
 * @polymerBehavior Polymer.IronValidatableBehavior
 * @appliesMixin ApiFormMixin
 */
class ApiHeadersForm extends mixinBehaviors([
    IronValidatableBehavior
  ], ApiFormMixin(PolymerElement)) {
  static get template() {
    return html`
    <style include="api-form-styles">
    :host {
      display: block;
      @apply --api-headers-form;
    }

    api-headers-form-item {
      @apply --layout-flex;
    }

    .delete-icon {
      margin-top: var(--api-headers-form-delete-icon-margin-top, 16px);
    }

    .enable-checkbox {
      margin-top: var(--api-headers-form-enable-checkbox-margin-top, 32px);
      margin-right: 8px;
    }

    :host([narrow]) .delete-icon {
      margin-top: var(--api-headers-form-delete-icon-margin-top-narrow, 16px);
    }

    :host([narrow]) .enable-checkbox {
      margin-top: var(--api-headers-form-enable-checkbox-margin-top-narrow, 32px);
    }

    [hidden] {
      display: none !important;
    }

    .empty-info {
      @apply --no-info-message;
    }
    </style>
    <arc-definitions id="definitions"></arc-definitions>
    <template is="dom-if" if="[[renderEmptyMessage]]">
      <p class="empty-info">Headers are not defined for this endpoint</p>
    </template>
    <iron-form>
      <form enctype="application/json">
        <template is="dom-if" if="[[renderOptionalCheckbox]]">
          <div class="optional-checkbox">
            <paper-checkbox class="toggle-checkbox" checked="{{optionalOpened}}" title="Shows or hides optional parameters">Show optional headers</paper-checkbox>
          </div>
        </template>
        <dom-repeat items="{{model}}">
          <template>
            <div class="form-item" data-optional\$="[[computeIsOptional(hasOptional, item)]]">
              <template is="dom-if" if="[[allowDisableParams]]">
                <paper-checkbox class="enable-checkbox" checked="{{item.schema.enabled}}" title="Enable or disable this header" disabled="[[readonly]]"></paper-checkbox>
              </template>
              <api-headers-form-item name="{{item.name}}" value="{{item.value}}" model="[[item]]" required\$="[[item.required]]" readonly="[[readonly]]" is-custom="[[item.schema.isCustom]]" is-array="[[item.schema.isArray]]" narrow="[[narrow]]" no-docs="[[noDocs]]">
                <paper-icon-button title="Removes header from the form" class="action-icon delete-icon" icon="arc:remove-circle-outline" on-tap="_removeCustom" slot="suffix" disabled="[[readonly]]"></paper-icon-button>
              </api-headers-form-item>
            </div>
          </template>
        </dom-repeat>
      </form>
    </iron-form>
    <template is="dom-if" if="[[allowCustom]]">
      <div class="add-action">
        <paper-button class="action-button" on-tap="add" title="Adds new header to the form" disabled="[[readonly]]">
          <iron-icon class="action-icon" icon="arc:add-circle-outline" alt="Add header icon"></iron-icon>
          Add header
        </paper-button>
      </div>
    </template>
`;
  }

  static get is() {
    return 'api-headers-form';
  }
  static get properties() {
    return {
      /**
       * Current value of the headers. Changing the value will update the list
       * of the headers.
       */
      value: {
        type: String,
        notify: true
      },
      /**
       * Prohibits rendering of the documentation (the icon and the
       * description).
       * Note, Set is separately for `api-view-model-transformer`
       * component as this only affects "custom" items.
       */
      noDocs: Boolean,
      /**
       * When set the editor is in read only mode.
       */
      readonly: Boolean,
      autoValidate: Boolean
    };
  }

  static get observers() {
    return [
      '_modelChanged(model.*, readonly)'
    ];
  }
  // Appends an empty header to the list.
  add() {
    if (!this.allowCustom) {
      return;
    }
    this.addCustom('header');
  }

  _modelChanged(record, readonly) {
    if (readonly) {
      return;
    }
    if (record.path === 'model' && this.invalid) {
      this.invalid = false;
    }
    if (['model', 'model.splices'].indexOf(record.path) !== -1) {
      this._updateValue();
      if (record.path === 'model' && record.value) {
        this._autoDescribeModel(record.value);
      }
    } else if (record.path.match(/model.\d+.(name|value|schema\.enabled)/)) {
      this._updateValue(this.autoValidate);
    }
  }
  /**
   * Updates value of the element when model change.
   * @param {Boolean} validate When true then it performs validation after setting
   * the value.
   */
  _updateValue(validate) {
    if (this.__updatingModelValue || this.readonly) {
      return;
    }
    this.__updatingModelValue = true;
    afterNextRender(this, () => {
      this.__updateValue(validate);
      this.__updatingModelValue = false;
    });
  }
  /**
   * Creates a header value for current model.
   * @param {Boolean} validate When true then it performs validation after setting
   * the value.
   */
  __updateValue(validate) {
    const h = this.model;
    if (!h || !h.length) {
      this.value = '';
      if (this.invalid) {
        this.invalid = false;
      }
      return;
    }
    let value = '';
    for (let i = 0, len = h.length; i < len; i++) {
      const item = h[i];
      if (item.schema && item.schema.enabled === false) {
        continue;
      }
      let n = item.name || '';
      let v = item.value || '';
      if (!n && !v) {
        continue;
      }
      if (v instanceof Array) {
        v = v.join(',');
      }
      if (!v && !item.required) {
        continue;
      }
      if (value[0]) {
        value += '\n';
      }
      value += n + ': ' + v;
    }
    this.set('value', value);
    if (validate) {
      this.validate();
    }
  }
  /**
   * Adds documentation for headers that doesn't have it already.
   *
   * @param {Array} model View model
   */
  _autoDescribeModel(model) {
    if (this.noDocs) {
      return;
    }
    model.forEach((item, index) => this._fillModelDescription(item, index));
  }
  /**
   * Queries for header information and updates header info if needed.
   *
   * @param {Object} item View model item
   * @param {Number} index Position of the item in `model` array
   */
  _fillModelDescription(item, index) {
    if (item.hasDescription || item.isCustom) {
      return;
    }
    const type = this._queryHeaderData(item.name);
    if (!type) {
      return;
    }
    this.set(['model', index, 'description'], type.desc);
    this.set(['model', index, 'hasDescription'], true);
    this.set(['model', index, 'schema', 'examples'], [type.example]);
  }

  /**
   * Queries for a header definition.
   *
   * @param {String} name header name to query
   * @return {Object|undefined} Header definition or undefined.
   */
  _queryHeaderData(name) {
    const suggestions = this.$.definitions.queryHeaders(name, 'request');
    if (!suggestions) {
      return;
    }
    name = name.toLowerCase();
    return suggestions.find((i) => i.key.toLowerCase() === name);
  }

  _getValidity() {
    const form = this.shadowRoot && this.shadowRoot.querySelector('iron-form');
    return form ? form.validate() : true;
  }
}

window.customElements.define(ApiHeadersForm.is, ApiHeadersForm);
